import logging
import requests
import urllib3
from requests.packages.urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("zeep").setLevel(logging.WARNING)
logging.getLogger("httpstream").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("rdflib").setLevel(logging.WARNING)
logging.getLogger("chardet").setLevel(logging.WARNING)



import os
import io
import yaml
import logging
import re
from datetime import timedelta
from importlib import import_module
from servicelayer.cache import make_key
from servicelayer.jobs import Dataset, Job
from servicelayer.extensions import get_entry_point

from memorious import settings
from memorious.core import conn, tags
from memorious.model import Crawl, Queue
from memorious.logic.stage import CrawlerStage

log = logging.getLogger(__name__)


class Crawler(object):
    """A processing graph that constitutes a crawler."""

    SCHEDULES = {
        "disabled": None,
        "hourly": timedelta(hours=1),
        "daily": timedelta(days=1),
        "weekly": timedelta(weeks=1),
        "monthly": timedelta(weeks=4),
    }

    def __init__(self, manager, source_file):
        self.manager = manager
        self.source_file = source_file
        with io.open(source_file, encoding="utf-8") as fh:
            self.config_yaml = fh.read()
            self.config = yaml.safe_load(self.config_yaml)

        self.name = os.path.basename(source_file)
        self.name = self.config.get("name", self.name)
        self.validate_name()
        self.description = self.config.get("description", self.name)
        self.category = self.config.get("category", "scrape")
        self.init_stage = self.config.get("init", "init")
        self.delay = int(self.config.get("delay", 0))
        self.expire = int(self.config.get("expire", settings.EXPIRE)) * 84600
        self.stealthy = self.config.get("stealthy", False)
        self.queue = Dataset(conn, self.name)
        self.aggregator_config = self.config.get("aggregator", {})

        self.stages = {}
        for name, stage in self.config.get("pipeline", {}).items():
            self.stages[name] = CrawlerStage(self, name, stage)

    def validate_name(self):
        if not re.match(r"^[A-Za-z0-9_-]+$", self.name):
            raise ValueError(
                "Invalid crawler name: %s. "
                "Allowed characters: A-Za-z0-9_-" % self.name
            )

    @property
    def aggregator_method(self):
        if self.aggregator_config:
            method = self.aggregator_config.get("method")
            if not method:
                return
            # method A: via a named Python entry point
            func = get_entry_point("memorious.operations", method)
            if func is not None:
                return func
            # method B: direct import from a module
            if ":" in method:
                package, method = method.rsplit(":", 1)
                module = import_module(package)
                return getattr(module, method)
            raise ValueError("Unknown method: %s", self.method_name)

    def aggregate(self, context):
        if self.aggregator_method:
            log.info("Running aggregator for %s" % self.name)
            params = self.aggregator_config.get("params", {})
            self.aggregator_method(context, params)

    def flush(self):
        """Delete all run-time data generated by this crawler."""
        self.queue.cancel()
        Crawl.flush(self)
        self.flush_tags()

    def flush_tags(self):
        tags.delete(prefix=make_key(self, "tag"))

    def cancel(self):
        Crawl.abort_all(self)
        self.queue.cancel()

    def run(self, incremental=None, run_id=None):
        """Queue the execution of a particular crawler."""
        state = {
            "crawler": self.name,
            "run_id": run_id or Job.random_id(),
            "incremental": settings.INCREMENTAL,
        }
        if incremental is not None:
            state["incremental"] = incremental

        # Cancel previous runs:
        self.cancel()
        init_stage = self.get(self.init_stage)
        Queue.queue(init_stage, state, {})

    @property
    def is_running(self):
        """Is the crawler currently running?"""
        for job in self.queue.get_jobs():
            if not job.is_done():
                return True
        return False

    @property
    def last_run(self):
        return Crawl.last_run(self)

    @property
    def op_count(self):
        """Total operations performed for this crawler"""
        return Crawl.op_count(self)

    @property
    def runs(self):
        return Crawl.runs(self)

    @property
    def latest_runid(self):
        return Crawl.latest_runid(self)

    @property
    def pending(self):
        status = self.queue.get_status()
        return status.get("pending")

    def get(self, name):
        return self.stages.get(name)

    def __str__(self):
        return self.name

    def __iter__(self):
        return iter(self.stages.values())

    def __repr__(self):
        return "<Crawler(%s)>" % self.name



import os
import pkg_resources
from servicelayer import env
from servicelayer import settings as sls

###############################################################################
# Core configuration
VERSION = pkg_resources.get_distribution("memorious").version
APP_NAME = env.get("MEMORIOUS_APP_NAME", "memorious")

# Enable debug logging etc.
DEBUG = env.to_bool("MEMORIOUS_DEBUG", default=False)
TESTING = False

# Base operating path
BASE_PATH = os.path.join(os.getcwd(), "data")
BASE_PATH = env.get("MEMORIOUS_BASE_PATH", BASE_PATH)

# Override servicelayer archive if undefined
sls.ARCHIVE_PATH = sls.ARCHIVE_PATH or os.path.join(BASE_PATH, "archive")

# Directory which contains crawler pipeline YAML specs
CONFIG_PATH = env.get("MEMORIOUS_CONFIG_PATH")

# Try and run scrapers in a way that only acquires new data
INCREMENTAL = env.to_bool("MEMORIOUS_INCREMENTAL", default=True)

# How many days until an incremental crawl expires
EXPIRE = env.to_int("MEMORIOUS_EXPIRE", 1)

# How many db inserts per minute
DB_RATE_LIMIT = env.to_int("MEMORIOUS_DB_RATE_LIMIT", 6000)

# How many http requests to a host per minute
HTTP_RATE_LIMIT = env.to_int("MEMORIOUS_HTTP_RATE_LIMIT", 120)

# Max number of tasks in a stage's task queue
MAX_QUEUE_LENGTH = env.to_int("MEMORIOUS_MAX_QUEUE_LENGTH", 50000)

# HTTP request configuration
HTTP_CACHE = env.to_bool("MEMORIOUS_HTTP_CACHE", default=True)

# HTTP request timeout
HTTP_TIMEOUT = float(env.to_int("MEMORIOUS_HTTP_TIMEOUT", 30))

# HTTP user agent default
USER_AGENT = "Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US; rv:1.1)"
USER_AGENT = "%s aleph.memorious/%s" % (USER_AGENT, VERSION)
USER_AGENT = env.get("MEMORIOUS_USER_AGENT", USER_AGENT)

# Datastore: operational data store (ODS) database connection
DATASTORE_FILE = os.path.join(BASE_PATH, "datastore.sqlite3")
DATASTORE_URI = "sqlite:///%s" % DATASTORE_FILE
DATASTORE_URI = env.get("MEMORIOUS_DATASTORE_URI", DATASTORE_URI)

# Tags cache table name
TAGS_TABLE = env.get("MEMORIOUS_TAGS_TABLE", "memorious_tags")



from pathlib import Path
from pprint import pprint  # noqa
from banal import clean_dict  # type: ignore

from servicelayer.cache import make_key  # type: ignore
from alephclient import settings
from alephclient.api import AlephAPI
from alephclient.util import backoff
from alephclient.errors import AlephException
from memorious.core import get_rate_limit  # type: ignore


def aleph_emit(context, data):
    api = get_api(context)
    if api is None:
        return
    collection_id = get_collection_id(context, api)
    content_hash = data.get("content_hash")
    source_url = data.get("source_url", data.get("url"))
    foreign_id = data.get("foreign_id", data.get("request_id", source_url))
    # Fetch document id from cache
    document_id = context.get_tag(make_key(collection_id, foreign_id, content_hash))
    if document_id:
        context.log.info("Skip aleph upload: %s", foreign_id)
        data["aleph_id"] = document_id
        context.emit(data=data, optional=True)
        return

    meta = {
        "crawler": context.crawler.name,
        "foreign_id": foreign_id,
        "source_url": source_url,
        "title": data.get("title"),
        "author": data.get("author"),
        "file_name": data.get("file_name"),
        "retrieved_at": data.get("retrieved_at"),
        "modified_at": data.get("modified_at"),
        "published_at": data.get("published_at"),
        "headers": data.get("headers", {}),
    }

    languages = context.params.get("languages")
    meta["languages"] = data.get("languages", languages)
    countries = context.params.get("countries")
    meta["countries"] = data.get("countries", countries)
    mime_type = context.params.get("mime_type")
    meta["mime_type"] = data.get("mime_type", mime_type)

    if data.get("aleph_folder_id"):
        meta["parent"] = {"id": data.get("aleph_folder_id")}

    meta = clean_dict(meta)
    # pprint(meta)
    label = meta.get("file_name", meta.get("source_url"))
    context.log.info("Upload: %s", label)
    with context.load_file(content_hash) as fh:
        if fh is None:
            return
        file_path = Path(fh.name).resolve()

        for try_number in range(api.retries):
            rate = settings.MEMORIOUS_RATE_LIMIT
            rate_limit = get_rate_limit("aleph", limit=rate)
            rate_limit.comply()
            try:
                res = api.ingest_upload(collection_id, file_path, meta)
                document_id = res.get("id")
                context.log.info("Aleph document entity ID: %s", document_id)
                # Save the document id in cache for future use
                context.set_tag(
                    make_key(collection_id, foreign_id, content_hash), document_id
                )
                data["aleph_id"] = document_id
                data["aleph_document"] = meta
                data["aleph_collection_id"] = collection_id
                context.emit(data=data, optional=True)
                return
            except AlephException as exc:
                if try_number > api.retries or not exc.transient:
                    context.emit_warning("Error: %s" % exc)
                    return
                backoff(exc, try_number)


def aleph_folder(context, data):
    api = get_api(context)
    if api is None:
        return
    collection_id = get_collection_id(context, api)
    foreign_id = data.get("foreign_id")
    if foreign_id is None:
        context.log.warning("No folder foreign ID!")
        return

    meta = {
        "crawler": context.crawler.name,
        "foreign_id": foreign_id,
        "source_url": data.get("source_url", data.get("url")),
        "title": data.get("title"),
        "file_name": data.get("file_name"),
        "retrieved_at": data.get("retrieved_at"),
        "modified_at": data.get("modified_at"),
        "published_at": data.get("published_at"),
    }

    if data.get("aleph_folder_id"):
        meta["parent"] = {"id": data.get("aleph_folder_id")}

    meta = clean_dict(meta)
    # pprint(meta)
    label = meta.get("file_name", meta.get("source_url"))
    context.log.info("Make folder: %s", label)
    for try_number in range(api.retries):
        rate = settings.MEMORIOUS_RATE_LIMIT
        rate_limit = get_rate_limit("aleph", limit=rate)
        rate_limit.comply()
        try:
            res = api.ingest_upload(collection_id, metadata=meta, sync=True)
            document_id = res.get("id")
            context.log.info("Aleph folder entity ID: %s", document_id)
            # Save the document id in cache for future use
            context.set_tag(make_key(collection_id, foreign_id), document_id)
            data["aleph_folder_id"] = document_id
            data["aleph_collection_id"] = collection_id
            context.emit(data=data, optional=True)
            return
        except AlephException as ae:
            if try_number > api.retries or not ae.transient:
                context.emit_warning("Error: %s" % ae)
                return
            backoff(ae, try_number)


def get_api(context):
    if not settings.HOST:
        context.log.warning("No $ALEPHCLIENT_HOST, skipping upload...")
        return None
    if not settings.API_KEY:
        context.log.warning("No $ALEPHCLIENT_API_KEY, skipping upload...")
        return None

    session_id = "memorious:%s" % context.crawler.name
    return AlephAPI(settings.HOST, settings.API_KEY, session_id=session_id)


def get_collection_id(context, api):
    if not hasattr(context.stage, "aleph_cid"):
        foreign_id = context.get("collection", context.crawler.name)
        config = {"label": context.crawler.description}
        collection = api.load_collection_by_foreign_id(foreign_id, config=config)
        context.stage.aleph_cid = collection.get("id")
    return context.stage.aleph_cid



from collections import namedtuple
import pytest
from memorious.helpers.rule import RULES, Rule, DomainRule, UrlPatternRule


spec = {
    "and": [
        {"domain": "occrp.org"},
        {
            "not": {
                "or": [
                    {"domain": "vis.occrp.org"},
                    {"domain": "tech.occrp.org"},
                    {"domain": "data.occrp.org"},
                    {"mime_group": "assets"},
                    {"mime_group": "images"},
                    {"pattern": "https://www.occrp.org/en/component/.*"},
                    {"pattern": "https://www.occrp.org/en/donate.*"},
                    {"pattern": "https://www.occrp.org/.*start=.*"},
                    {"pattern": "https://www.occrp.org/ru/.*"},
                ]
            }
        },
    ]
}

invalid_spec = {
    "and": [{"domain": "occrp.org"}, {"not": {"domain": "vis.occrp.org"}}],
    "not": {"mime_group": "images"},
}

rule = {}


class TestRule(object):
    def test_get_rule(self):
        with pytest.raises(Exception):
            Rule.get_rule(invalid_spec)
        assert isinstance(Rule.get_rule(spec), RULES["and"])


class TestDomainRule(object):
    def test_domain_rule(self):
        rule = DomainRule("occrp.org")
        rule.configure()
        Response = namedtuple("Response", "url")
        res = Response(url="http://occrp.org")
        assert rule.apply(res)
        res = Response(url="http://not-occrp.org")
        assert rule.apply(res) is False


class TestUrlPatternRule(object):
    def test_url_pattern(self):
        rule = UrlPatternRule("https://www.occrp.org/en/donate.*")
        rule.configure()
        Response = namedtuple("Response", "url")
        res = Response(url="https://www.occrp.org/en/donate.html")
        assert rule.apply(res)
        res = Response(url="http://not-occrp.org")
        assert rule.apply(res) is False



from urllib.parse import urlparse
from requests.exceptions import RequestException
from servicelayer.cache import make_key

from memorious.helpers.rule import Rule


def fetch(context, data):
    """Do an HTTP GET on the ``url`` specified in the inbound data."""
    url = data.get("url")
    if urlparse(url).scheme not in ("http", "https", ""):
        context.log.info("Fetch skipped. Unsupported scheme: %r", url)
        return
    attempt = data.pop("retry_attempt", 1)
    try:
        result = context.http.get(url, lazy=True)
        rules = context.get("rules", {"match_all": {}})
        if not Rule.get_rule(rules).apply(result):
            context.log.info("Fetch skip: %r", result.url)
            return

        if not result.ok:
            err = (result.url, result.status_code)
            context.emit_warning("Fetch fail [%s]: HTTP %s" % err)
            if not context.params.get("emit_errors", False):
                return
        else:
            context.log.info("Fetched [%s]: %r", result.status_code, result.url)

        data.update(result.serialize())
        if url != result.url:
            tag = make_key(context.run_id, url)
            context.set_tag(tag, None)
        context.emit(data=data)
    except RequestException as ce:
        retries = int(context.get("retry", 3))
        if retries >= attempt:
            context.log.warn("Retry: %s (error: %s)", url, ce)
            data["retry_attempt"] = attempt + 1
            context.recurse(data=data, delay=2 ** attempt)
        else:
            context.emit_warning("Fetch fail [%s]: %s" % (url, ce))


def session(context, data):
    """Set some HTTP parameters for all subsequent requests.

    This includes ``user`` and ``password`` for HTTP basic authentication,
    and ``user_agent`` as a header.
    """
    context.http.reset()

    user = context.get("user")
    password = context.get("password")

    if user is not None and password is not None:
        context.http.session.auth = (user, password)

    user_agent = context.get("user_agent")
    if user_agent is not None:
        context.http.session.headers["User-Agent"] = user_agent

    referer = context.get("url")
    if referer is not None:
        context.http.session.headers["Referer"] = referer

    proxy = context.get("proxy")
    if proxy is not None:
        proxies = {"http": proxy, "https": proxy}
        context.http.session.proxies = proxies

    # Explicitly save the session because no actual HTTP requests were made.
    context.http.save()
    context.emit(data=data)



from banal import ensure_list
from ftmstore import get_dataset as get_ftmstore_dataset
from ftmstore.settings import DATABASE_URI, DEFAULT_DATABASE_URI

from memorious.settings import DATASTORE_URI
from memorious.operations.aleph import get_api

ORIGIN = "memorious"


def get_dataset(context, origin=ORIGIN):
    name = context.get("dataset", context.crawler.name)
    origin = context.get("dataset", origin)
    # Either use a database URI that has been explicitly set as a
    # backend, or default to the memorious datastore.
    database_uri = DATABASE_URI
    if DATABASE_URI == DEFAULT_DATABASE_URI:
        database_uri = DATASTORE_URI
    return get_ftmstore_dataset(name, database_uri=database_uri, origin=origin)


def ftm_store(context, data):
    """Store an entity or a list of entities to an ftm store."""
    # This is a simplistic implementation of a balkhash memorious operation.
    # It is meant to serve the use of OCCRP where we pipe data into postgresql.
    dataset = get_dataset(context)
    bulk = dataset.bulk()
    entities = ensure_list(data.get("entities", data))
    for entity in entities:
        context.log.debug("Store entity [%(schema)s]: %(id)s", entity)
        bulk.put(entity, entity.pop("fragment", None))
        context.emit(rule="fragment", data=data, optional=True)
    context.emit(data=data, optional=True)
    bulk.flush()


def ftm_load_aleph(context, data):
    """Write each entity from an ftm store to Aleph via the _bulk API."""
    api = get_api(context)
    if api is None:
        return
    foreign_id = context.params.get("foreign_id", context.crawler.name)
    collection = api.load_collection_by_foreign_id(foreign_id, {})
    collection_id = collection.get("id")
    unsafe = context.params.get("unsafe", False)
    entities = get_dataset(context)
    api.write_entities(collection_id, entities, unsafe=unsafe)



from datetime import datetime
from urllib.parse import urljoin

API_HOST = "https://api.www.documentcloud.org"
ASSET_HOST = "https://assets.documentcloud.org"
DOCUMENT_HOST = "https://www.documentcloud.org"
DEFAULT_INSTANCE = "documentcloud"

LANGUAGES = {
    "ara": "ar",
    "zho": "zh",
    "tra": "zh",
    "hrv": "hr",
    "dan": "da",
    "nld": "nl",
    "eng": "en",
    "fra": "fr",
    "deu": "de",
    "heb": "he",
    "hun": "hu",
    "ind": "id",
    "ita": "it",
    "jpn": "ja",
    "kor": "ko",
    "nor": "no",
    "por": "pt",
    "ron": "ro",
    "rus": "ru",
    "spa": "es",
    "swe": "sv",
    "ukr": "uk",
}


def documentcloud_query(context, data):
    host = context.get("host", API_HOST)
    instance = context.get("instance", DEFAULT_INSTANCE)
    query = context.get("query")
    page = data.get("page", 1)

    search_url = urljoin(host, "/api/documents/search")

    res = context.http.get(
        search_url,
        params={"q": query, "per_page": 100, "page": page, "expand": "organization"},
    )

    documents = res.json.get("results", [])

    for document in documents:
        doc = {
            "foreign_id": "%s:%s" % (instance, document.get("id")),
            "url": "{}/documents/{}/{}.pdf".format(
                ASSET_HOST, document.get("id"), document.get("slug")
            ),
            "source_url": "{}/documents/{}-{}".format(
                DOCUMENT_HOST, document.get("id"), document.get("slug")
            ),
            "title": document.get("title"),
            "author": document.get("organization", {}).get("name"),
            "file_name": "{}.pdf".format(document.get("slug")),
            "mime_type": "application/pdf",
        }

        lang = LANGUAGES.get(document.get("language"))
        if lang is not None:
            data["languages"] = [lang]

        published = document.get("created_at")
        if published is not None:
            dt = datetime.strptime(published, "%Y-%m-%dT%H:%M:%S.%fZ")
            doc["published_at"] = dt.isoformat()

        context.emit(data=doc)

    if len(documents):
        context.recurse(data={"page": page + 1})



WEB = [
    "text/html",
    "text/plain",
    "text/xml",
    "application/xml",
]

IMAGES = [
    "image/jpeg",
    "image/bmp",
    "image/png",
    "image/tiff",
    "image/gif",
    "application/postscript",
    "image/vnd.dxf",
    "image/svg+xml",
    "image/x-pict",
    "image/x-ms-bmp",
]

MEDIA = [
    "audio/mpeg",
    "video/x-flv",
    "video/mp4",
    "video/mp2t",
    "video/3gpp",
    "video/quicktime",
    "video/x-msvideo",
    "video/x-ms-wmv",
]

DOCUMENTS = [
    "application/vnd.ms-excel",
    "application/msword",
    "application/pdf",
    ".pdf",  # seen in the wild (TM)
    "application/vnd.ms-powerpoint",
    "text/richtext",
    "application/vnd.oasis.opendocument.text",
    "application/rtf",
    "application/vnd.oasis.opendocument.spreadsheet",
    "application/x-rtf",
    "application/vnd.oasis.opendocument.graphics",
    "application/vnd.oasis.opendocument.presentation",
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",  # noqa
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document",  # noqa
    "application/vnd.openxmlformats-officedocument.presentationml.presentation",  # noqa
]

ARCHIVES = [
    "application/zip",
    "application/x-rar-compressed",
    "application/x-tar",
    "application/x-gzip",
    "application/x-7z-compressed",
]

ASSETS = [
    "text/css",
    "application/javascript",
    "application/json",
    "image/x-icon",
    "application/rss+xml",
    "application/atom+xml",
    "application/opensearchdescription+xml",
]

GROUPS = {
    "web": WEB,
    "images": IMAGES,
    "media": MEDIA,
    "documents": DOCUMENTS,
    "archives": ARCHIVES,
    "assets": ASSETS,
}

KNOWN = WEB + IMAGES + MEDIA + DOCUMENTS + ARCHIVES + ASSETS
NON_HTML = [t for t in KNOWN if t not in WEB]



from banal import ensure_list
from datetime import datetime, timedelta


def seed(context, data):
    """Initialize a crawler with a set of seed URLs.

    The URLs are given as a list or single value to the ``urls`` parameter.

    If this is called as a second stage in a crawler, the URL will be formatted
    against the supplied ``data`` values, e.g.:

        https://crawl.site/entries/%(number)s.html
    """
    for key in ("url", "urls"):
        for url in ensure_list(context.params.get(key)):
            url = url % data
            context.emit(data={"url": url})


def enumerate(context, data):
    """Iterate through a set of items and emit each one of them."""
    items = ensure_list(context.params.get("items"))
    for item in items:
        data["item"] = item
        context.emit(data=data)


def tee(context, data):
    """Trigger multiple subsequent stages in parallel."""
    for rule, _ in context.stage.handlers.items():
        context.emit(rule=rule, data=data)


def sequence(context, data):
    """Generate a sequence of numbers.

    It is the memorious equivalent of the xrange function, accepting the
    ``start``, ``stop`` and ``step`` parameters.

    This can run in two ways:
    * As a single function generating all numbers in the given range.
    * Recursively, generating numbers one by one with an optional ``delay``.

    The latter mode is useful in order to generate very large sequences
    without completely clogging up the user queue.

    If an optional ``tag`` is given, each number will be emitted only once
    across multiple runs of the crawler.
    """
    number = data.get("number", context.params.get("start", 1))
    stop = context.params.get("stop")
    step = context.params.get("step", 1)
    delay = context.params.get("delay")
    prefix = context.params.get("tag")
    while True:
        tag = None if prefix is None else "%s:%s" % (prefix, number)

        if tag is None or not context.check_tag(tag):
            data["number"] = number
            context.emit(data=data)

        if tag is not None:
            context.set_tag(tag, True)

        number = number + step
        if step > 0 and number >= stop:
            break
        if step < 0 and number <= stop:
            break

        if delay is not None:
            data["number"] = number
            context.recurse(data=data, delay=delay)
            break


def dates(context, data):
    format = context.params.get("format", "%Y-%m-%d")
    delta = timedelta(
        days=context.params.get("days", 0), weeks=context.params.get("weeks", 0)
    )
    if delta == timedelta():
        context.log.error("No interval given: %r", context.params)
        return

    if "end" in context.params:
        current = context.params.get("end")
        current = datetime.strptime(current, format)
    else:
        current = datetime.utcnow()

    if "current" in data:
        current = datetime.strptime(data.get("current"), format)

    if "begin" in context.params:
        begin = context.params.get("begin")
        begin = datetime.strptime(begin, format)
    else:
        steps = context.params.get("steps", 100)
        begin = current - (delta * steps)

    context.emit(
        data={"date": current.strftime(format), "date_iso": current.isoformat()}
    )
    current = current - delta
    if current >= begin:
        data = {"current": current.strftime(format)}
        context.recurse(data=data)



import click
import logging
from tabulate import tabulate

from memorious import settings
from memorious.core import manager, init_memorious, conn
from memorious.worker import get_worker

log = logging.getLogger(__name__)


@click.group()
@click.option("--debug/--no-debug", default=False, envvar="MEMORIOUS_DEBUG")
@click.option("--cache/--no-cache", default=True, envvar="MEMORIOUS_HTTP_CACHE")
@click.option(
    "--incremental/--non-incremental", default=True, envvar="MEMORIOUS_INCREMENTAL"
)
def cli(debug, cache, incremental):
    """Crawler framework for documents and structured scrapers."""
    settings.HTTP_CACHE = cache
    settings.INCREMENTAL = incremental
    settings.DEBUG = debug
    init_memorious()


def get_crawler(name):
    crawler = manager.get(name)
    if crawler is None:
        msg = "Crawler [%s] not found." % name
        raise click.BadParameter(msg, param=crawler)
    return crawler


@cli.command("run")
@click.argument("crawler")
@click.option("--threads", type=int, default=None)
@click.option("--flush", is_flag=True, default=False)
@click.option("--flushall", is_flag=True, default=False)
def run(crawler, threads=None, flush=False, flushall=False):
    """Run a specified crawler in synchronous mode."""
    crawler = get_crawler(crawler)
    settings._crawler = crawler
    if flush:
        crawler.flush()
    if flushall:
        conn.flushall()
    crawler.run()
    if threads is not None and threads > 1:
        if settings.sls.REDIS_URL is None:
            log.warning(
                "REDIS_URL not set. Can't run in multithreaded mode without Redis. Exiting."
            )
            return
        if settings.DATASTORE_URI.startswith("sqlite:///"):
            log.warning(
                "Can't run in multithreaded mode with sqlite database. Exiting."
            )
            return
        worker = get_worker(num_threads=threads)  # multithreaded worker
        worker.run()
    else:
        worker = get_worker()
        worker.sync()


@cli.command()
@click.argument("crawler")
def cancel(crawler):
    """Abort execution of a specified crawler."""
    crawler = get_crawler(crawler)
    crawler.cancel()


@cli.command()
@click.argument("crawler")
def flush(crawler):
    """Delete all data generated by a crawler."""
    crawler = get_crawler(crawler)
    crawler.flush()


@cli.command("flush-tags")
@click.argument("crawler")
def flush_tags(crawler):
    """Delete all tags generated by a crawler."""
    crawler = get_crawler(crawler)
    crawler.flush_tags()


@cli.command("list")
def index():
    """List the available crawlers."""
    crawler_list = []
    for crawler in manager:
        crawler_list.append(
            [
                crawler.name,
                crawler.description,
            ]
        )
    headers = [
        "Name",
        "Description",
    ]
    print(tabulate(crawler_list, headers=headers))


@cli.command("status")
@click.argument("crawler")
def status(crawler):
    """Status of a crawler."""
    crawler = get_crawler(crawler)
    prop_list = []
    last_run = crawler.last_run
    if last_run:
        last_run = last_run.isoformat() + " UTC"
    prop_list.append(
        [
            crawler.name,
            crawler.description,
            crawler.is_running,
            last_run,
            crawler.op_count,
            crawler.pending,
        ]
    )
    headers = [
        "Name",
        "Description",
        "Running?",
        "Last Active",
        "Op Count",
        "Pending Ops",
    ]
    print(tabulate(prop_list, headers=headers))


@cli.command()
def killthekitten():
    """Completely kill redis contents."""
    from memorious.core import connect_redis

    conn = connect_redis()
    conn.flushall()


def main():
    cli(obj={})


if __name__ == "__main__":
    main()



import structlog

from servicelayer.worker import Worker
from servicelayer.logs import apply_task_context

from memorious import settings
from memorious.logic.context import Context
from memorious.logic.stage import CrawlerStage
from memorious.core import conn, crawler

log = structlog.get_logger(__name__)


class MemoriousWorker(Worker):
    def handle(self, task):
        apply_task_context(task)
        data = task.payload
        stage = CrawlerStage.detach_namespace(task.stage.stage)
        state = task.context
        context = Context.from_state(state, stage)
        context.execute(data)

    def after_task(self, task):
        if task.job.is_done():
            stage = CrawlerStage.detach_namespace(task.stage.stage)
            state = task.context
            context = Context.from_state(state, stage)
            context.crawler.aggregate(context)
            self.shutdown()

    def get_stages(self):
        return [stage.namespaced_name for stage in crawler.stages.values()]


def get_worker(num_threads=None):
    return MemoriousWorker(
        conn=conn,
        num_threads=num_threads or settings.sls.WORKER_THREADS,
    )



from pathlib import Path
from urllib.parse import urljoin, urlparse, unquote


def _get_url_file_name(url):
    path = urlparse(url).path
    try:
        path = unquote(path)
    except (TypeError, ValueError):
        pass
    return Path(path).name


def dav_index(context, data):
    """List files in a WebDAV directory."""
    # This is made to work with ownCloud/nextCloud, but some rumor has
    # it they are "standards compliant" and it should thus work for
    # other DAV servers.
    url = data.get("url")
    context.log.info("Fetching WebDAV path: %s" % url)
    result = context.http.request("PROPFIND", url)
    for resp in result.xml.findall("./{DAV:}response"):
        href = resp.findtext("./{DAV:}href")
        if href is None:
            continue

        child_url = urljoin(url, href)
        if child_url == url:
            continue
        child = dict(data)
        child["url"] = child_url
        child["foreign_id"] = child_url
        child["file_name"] = _get_url_file_name(href)

        rule = "file"
        if resp.find(".//{DAV:}collection") is not None:
            rule = "folder"
        context.emit(data=child, rule=rule)



from setuptools import setup, find_packages

setup(
    name="memorious",
    version="2.0.0",
    description="A minimalistic, recursive web crawling library for Python.",
    long_description="",
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    keywords="",
    author="Organized Crime and Corruption Reporting Project",
    author_email="data@occrp.org",
    url="http://github.com/alephdata/memorious",
    license="MIT",
    packages=find_packages(exclude=["ez_setup", "examples", "tests"]),
    namespace_packages=[],
    include_package_data=True,
    package_data={
        "memorious": [
            "migrate/*",
            "migrate/versions/*",
            "migrate/ui/static/*",
            "migrate/ui/templates/*",
        ]
    },
    zip_safe=False,
    install_requires=[
        "banal >= 1.0.1, < 2.0.0",
        "click",
        "requests[security] >= 2.21.0",
        "PySocks == 1.7.1",
        "requests_ftp",
        "lxml >= 4",
        "PyYAML == 5.4.1",
        "normality >= 2.1.1, < 3.0.0",
        "tabulate",
        "python-dateutil == 2.8.1",
        "dataset >= 1.0.8",
        "servicelayer[google,amazon] == 1.17.2",
        "pantomime == 0.4.1",
        "alephclient >= 2.2.2",
        "followthemoney >= 2.3.1",
        "followthemoney-store >= 3.0.1",
        "dateparser",
        "stringcase",
        "flask",
        "babel",
    ],
    entry_points={
        "console_scripts": ["memorious = memorious.cli:main"],
        "memorious.operations": [
            "memorious = memorious.cli:main",
            "fetch = memorious.operations.fetch:fetch",
            "session = memorious.operations.fetch:session",
            "dav_index = memorious.operations.webdav:dav_index",
            "parse = memorious.operations.parse:parse",
            "clean_html = memorious.operations.clean:clean_html",
            "seed = memorious.operations.initializers:seed",
            "tee = memorious.operations.initializers:tee",
            "sequence = memorious.operations.initializers:sequence",
            "dates = memorious.operations.initializers:dates",
            "enumerate = memorious.operations.initializers:enumerate",
            "inspect = memorious.operations.debug:inspect",
            "documentcloud_query = memorious.operations.documentcloud:documentcloud_query",  # noqa
            "directory = memorious.operations.store:directory",
            "extract = memorious.operations.extract:extract",
            "db = memorious.operations.db:db",
            "ftp_fetch = memorious.operations.ftp:ftp_fetch",
            "aleph_emit = memorious.operations.aleph:aleph_emit",
            "aleph_folder = memorious.operations.aleph:aleph_folder",
            "balkhash_put = memorious.operations.ftm:ftm_store",
            "ftm_store = memorious.operations.ftm:ftm_store",
            "ftm_load_aleph = memorious.operations.ftm:ftm_load_aleph",
        ],
    },
    extras_require={
        "dev": [
            "pytest",
            "pytest-env",
            "pytest-cov",
            "pytest-mock",
            "sphinx",
            "sphinx_rtd_theme",
            "recommonmark",
        ],
        "ocr": [
            "tesserocr",
        ],
    },
)
